@@ -14,16 +14,18 @@
  * limitations under the License.
  */
 
-#include "mongoc/mongoc-array-private.h"
-#include "mongoc/mongoc-error.h"
-#include "mongoc/mongoc-server-description-private.h"
-#include "mongoc/mongoc-topology-description-apm-private.h"
-#include "mongoc/mongoc-trace-private.h"
-#include "mongoc/mongoc-util-private.h"
-#include "mongoc/mongoc-read-prefs-private.h"
-#include "mongoc/mongoc-set-private.h"
-#include "mongoc/mongoc-client-private.h"
-#include "mongoc/mongoc-thread-private.h"
+#include "mongoc-array-private.h"
+#include "mongoc-error.h"
+#include "mongoc-server-description-private.h"
+#include "mongoc-topology-description-apm-private.h"
+#include "mongoc-trace-private.h"
+#include "mongoc-util-private.h"
+#include "mongoc-read-prefs-private.h"
+#include "mongoc-set-private.h"
+#include "mongoc-client-private.h"
+#include "mongoc-thread-private.h"
+#include "mongoc-host-list-private.h"
+#include "utlist.h"
 
 
 static bool
@@ -1130,7 +1132,7 @@ mongoc_topology_description_invalidate_server (
    BSON_ASSERT (error);
 
    /* send NULL ismaster reply */
-   mongoc_topology_description_handle_ismaster (topology, id, NULL, 0, error);
+   mongoc_topology_description_handle_ismaster (topology, id, NULL, MONGOC_RTT_UNSET, error);
 }
 
 /*
@@ -1903,8 +1905,8 @@ _mongoc_topology_description_check_compatible (
  * mongoc_topology_description_handle_ismaster --
  *
  *      Handle an ismaster. This is called by the background SDAM process,
- *      and by client when invalidating servers. If there was an error
- *      calling ismaster, pass it in as @error.
+ *      and by client when performing a handshake or invalidating servers.
+ *      If there was an error calling ismaster, pass it in as @error.
  *
  *      NOTE: this method should only be called while holding the mutex on
  *      the owning topology object.
@@ -1923,6 +1925,10 @@ mongoc_topology_description_handle_ismaster (
    mongoc_topology_description_t *prev_td = NULL;
    mongoc_server_description_t *prev_sd = NULL;
    mongoc_server_description_t *sd;
+   bson_iter_t iter;
+   /* sd_changed is set if the server description meaningfully changed AND
+    * callbacks are registered. */
+   bool sd_changed = false;
 
    BSON_ASSERT (topology);
    BSON_ASSERT (server_id != 0);
@@ -1937,25 +1943,90 @@ mongoc_topology_description_handle_ismaster (
       _mongoc_topology_description_copy_to (topology, prev_td);
    }
 
-   if (topology->apm_callbacks.server_changed) {
+   if (ismaster_response &&
+       bson_iter_init_find (&iter, ismaster_response, "topologyVersion") &&
+       BSON_ITER_HOLDS_DOCUMENT (&iter)) {
+      bson_t incoming_topology_version;
+      const uint8_t *bytes;
+      uint32_t len;
+
+      bson_iter_document (&iter, &len, &bytes);
+      bson_init_static (&incoming_topology_version, bytes, len);
+
+      if (mongoc_server_description_topology_version_cmp (
+             &sd->topology_version, &incoming_topology_version) == 1) {
+         TRACE ("%s", "topology version is strictly less. Skipping.");
+         if (prev_td) {
+            mongoc_topology_description_destroy (prev_td);
+            bson_free (prev_td);
+         }
+         return;
+      }
+   }
+
+   if (topology->apm_callbacks.topology_changed ||
+       topology->apm_callbacks.server_changed) {
+      /* Only copy the previous server description if a monitoring callback is
+       * registered. */
       prev_sd = mongoc_server_description_new_copy (sd);
    }
 
+   DUMP_BSON (ismaster_response);
    /* pass the current error in */
+
    mongoc_server_description_handle_ismaster (
       sd, ismaster_response, rtt_msec, error);
 
+   /* if the user specified a set_name in the connection string
+    * and they are in topology type single, check that the set name
+    * matches. */
+   if (topology->set_name && topology->type == MONGOC_TOPOLOGY_SINGLE) {
+      bool wrong_set_name = false;
+      bson_error_t set_name_err = {0};
+
+      if (!sd->set_name) {
+         wrong_set_name = true;
+         bson_set_error (&set_name_err,
+                         MONGOC_ERROR_SERVER_SELECTION,
+                         MONGOC_ERROR_SERVER_SELECTION_FAILURE,
+                         "no reported set name, but expected '%s'",
+                         topology->set_name);
+      } else if (0 != strcmp (sd->set_name, topology->set_name)) {
+         wrong_set_name = true;
+         bson_set_error (&set_name_err,
+                         MONGOC_ERROR_SERVER_SELECTION,
+                         MONGOC_ERROR_SERVER_SELECTION_FAILURE,
+                         "reported set name '%s' does not match '%s'",
+                         sd->set_name,
+                         topology->set_name);
+      }
+
+      if (wrong_set_name) {
+         /* Replace with unknown. */
+         TRACE ("%s", "wrong set name");
+         mongoc_server_description_handle_ismaster (
+            sd, NULL, MONGOC_RTT_UNSET, &set_name_err);
+      }
+   }
+
    mongoc_topology_description_update_cluster_time (topology,
                                                     ismaster_response);
-   _mongoc_topology_description_monitor_server_changed (topology, prev_sd, sd);
+
+   if (prev_sd) {
+      sd_changed = !_mongoc_server_description_equal (prev_sd, sd);
+   }
+   if (sd_changed) {
+      _mongoc_topology_description_monitor_server_changed (
+         topology, prev_sd, sd);
+   }
 
    if (gSDAMTransitionTable[sd->type][topology->type]) {
-      TRACE ("Transitioning to %s for %s",
+      TRACE ("Topology description %s handling server description %s",
              _mongoc_topology_description_type (topology),
              mongoc_server_description_type (sd));
       gSDAMTransitionTable[sd->type][topology->type](topology, sd);
    } else {
-      TRACE ("No transition entry to %s for %s",
+      TRACE ("Topology description %s ignoring server description %s",
              _mongoc_topology_description_type (topology),
              mongoc_server_description_type (sd));
    }
@@ -1966,16 +2037,19 @@ mongoc_topology_description_handle_ismaster (
    if (ismaster_response && (!error || !error->code)) {
       _mongoc_topology_description_check_compatible (topology);
    }
-   _mongoc_topology_description_monitor_changed (prev_td, topology);
+
+   /* If server description did not change, then neither did topology
+    * description */
+   if (sd_changed) {
+      _mongoc_topology_description_monitor_changed (prev_td, topology);
+   }
 
    if (prev_td) {
       mongoc_topology_description_destroy (prev_td);
       bson_free (prev_td);
    }
 
-   if (prev_sd) {
-      mongoc_server_description_destroy (prev_sd);
-   }
+   mongoc_server_description_destroy (prev_sd);
 }
 
 /*
@@ -2119,3 +2193,46 @@ mongoc_topology_description_get_servers (
 
    return sds;
 }
+
+typedef struct {
+   mongoc_host_list_t *host_list;
+   mongoc_topology_description_t *td;
+} _remove_if_not_in_host_list_ctx_t;
+
+bool
+_remove_if_not_in_host_list_cb (void *sd_void, void *ctx_void)
+{
+   _remove_if_not_in_host_list_ctx_t *ctx;
+   mongoc_topology_description_t *td;
+   mongoc_server_description_t *sd;
+   mongoc_host_list_t *host_list;
+
+   ctx = ctx_void;
+   sd = sd_void;
+   host_list = ctx->host_list;
+   td = ctx->td;
+
+   if (_mongoc_host_list_contains_one (host_list, &sd->host)) {
+      return true;
+   }
+   _mongoc_topology_description_remove_server (td, sd);
+   return true;
+}
+
+void
+mongoc_topology_description_reconcile (mongoc_topology_description_t *td,
+                                       mongoc_host_list_t *host_list)
+{
+   mongoc_host_list_t *host;
+   _remove_if_not_in_host_list_ctx_t ctx;
+
+   LL_FOREACH (host_list, host)
+   {
+      /* "add" is really an "upsert" */
+      mongoc_topology_description_add_server (td, host->host_and_port, NULL);
+   }
+
+   ctx.host_list = host_list;
+   ctx.td = td;
+   mongoc_set_for_each (td->servers, _remove_if_not_in_host_list_cb, &ctx);
+}
